/******************************************************************************
 * The copyright in this software is being made available under the BSD        *
 * License, included below. This software may be subject to other third party  *
 * and contributor rights, including patent rights, and no such rights are     *
 * granted under this license.                                                 *
 *    <OWNER> = ISO                                                            *
 *    <ORGANIZATION> = ISO                                                     *
 *    <YEAR> = 2015                                                            *
 *    Copyright (c) 2012-2015, ISO                                             *
 *                                                                             *
 * All rights reserved.                                                        *
 * Redistribution and use in source and binary forms, with or without          *
 * modification, are permitted provided that the following conditions are met: *
 *       * Redistributions of source code must retain the above copyright      *
 *         notice,this list of conditions and the following disclaimer.        *
 *       * Redistributions in binary form must reproduce the above             *
 *         copyright notice, this list of conditions and the following         *
 *         disclaimer in the documentation and/or other materials provided     *
 *         with the distribution.                                              *
 *       * Neither the name of the <ORGANIZATION> nor the names of its         *
 *         contributors may be used to endorse or promote products derived     *
 *         from this software without specific prior written permission.       *
 *                                                                             *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" *
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE   *
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE  *
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE   *
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR         *
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF        *
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS    *
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN     *
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)     *
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE  *
 * POSSIBILITY OF SUCH DAMAGE.                                                 *
 ******************************************************************************/

/** @file
 *  @brief Header file containing declarations of functions and data structures for the JPEG interface.
 *
 *  This file contains declarations of the functions and data structure that are used by the JPEG interface.
 */
 
#ifndef JH_JPEG_WRAP_H
#define JH_JPEG_WRAP_H
#include "jh_datatypes.h"

#ifdef __cplusplus
extern "C" {
#endif

/** @defgroup jpegint JPEG Interface
 *  @brief Defines the API for the JPEG interface.
 *  @{
 */
 
/** @brief Compresses a standard dynamic range image (SDRI) in JPEG JFIF file format, attaches APP11 segments and an ICC profile,
 *  and stores the resulting image in memory.
 *  @return 0 if the function does not succeed.
 */
int jh_jpeg_compress_and_attach(
            const jh_image_t *sdri,    /**< [in]     Input standard dynamic range image (SDRI), also known as the tone-mapped
                                                     image, or base image. */
            int qual,                  /**< [in]     Encoding quality to compress the input SDRI. Valid values 0 to 100. */
            int subsampling,           /**< [in]     Subsampling. Valid values 0, 420 (the same as 0), 422, 444 */
            jh_marker_list_t *markers, /**< [in,out] User-defined container for storing JPEG APP and COM data. Can be NULL. */
            jh_stream_t* stream,       /**< [in,out] User-implemented stream to hold encoded data. */
            jh_monitor_t *monitor,     /**< [in]     Progress status of the function. Initial value of the jh_monitor_t
                                                     @c cancel field must be 0. */
            jh_error_t* error          /**< [in,out] User-allocated error-reporting structure that contains information about the
                                                     error. Before calling this function, the value of @c error should be set to 0. */
);

/** @brief Reads the JPEG-HDR file to get basic information about the image.
 *  @return 0 if the function does not succeed.
 */
int jh_jpeg_read_header(
            jh_stream_t*  stream,       /**< [in]     Input JPEG byte stream. */
            int* width,                 /**< [out]    Number of columns in the image. */
            int* height,                /**< [out]    Number of rows in the image. */
            int* channels,              /**< [out]    Number of channels in the image. */
            jh_error_t* error           /**< [in,out] User-allocated error-reporting structure that contains information about the
                                                      error. Before calling this function, the value of @c error should be set to 0. */
);

/** @brief Extracts markers from the JPEG stream and stores them in the marker list.
 *  @return 0 if the function does not succeed.
 */
int jh_jpeg_extract_markers(
        jh_stream_t*  stream,           /**< [in]     Input JPEG byte stream. */
        jh_marker_list_t* markers,      /**< [in,out] User-defined marker container. */
        jh_error_t *error               /**< [in,out] User-allocated error-reporting structure that contains information about
                                                      the error, if any. */
);

/** @brief Decodes the image from the JPEG stream.
 *  @return 0 if the function does not succeed.
 */
int jh_jpeg_extract_image(
        jh_stream_t* stream,            /**< [in]     Input JPEG byte stream. */
        int scale_num,                  /**< [in]     JPEG numerator parameter. */
        int scale_denom,                /**< [in]     JPEG denominator parameter. @c scale_num and @c scale_denom scale down the
                                                      resulting JPEG image. */
        jh_image_t* sdri,               /**< [out]    Decoded image. */
        jh_error_t* error               /**< [in,out] User-allocated error-reporting structure that contains information about
                                                      the error, if any. */
        );

/** @brief Releases the image data buffer.
 */
void jh_jpeg_release_image_data(
        unsigned char* data             /**< [in]     Image data allocated by jh_jpeg_extract_image(). */
        );

/**@}*/  /* end of defgroup jpegint */


#ifdef __cplusplus
}
#endif
#endif /*JH_JPEG_WRAP_H*/

