/******************************************************************************
 * The copyright in this software is being made available under the BSD        *
 * License, included below. This software may be subject to other third party  *
 * and contributor rights, including patent rights, and no such rights are     *
 * granted under this license.                                                 *
 *    <OWNER> = ISO                                                            *
 *    <ORGANIZATION> = ISO                                                     *
 *    <YEAR> = 2015                                                            *
 *    Copyright (c) 2012-2015, ISO                                             *
 *                                                                             *
 * All rights reserved.                                                        *
 * Redistribution and use in source and binary forms, with or without          *
 * modification, are permitted provided that the following conditions are met: *
 *       * Redistributions of source code must retain the above copyright      *
 *         notice,this list of conditions and the following disclaimer.        *
 *       * Redistributions in binary form must reproduce the above             *
 *         copyright notice, this list of conditions and the following         *
 *         disclaimer in the documentation and/or other materials provided     *
 *         with the distribution.                                              *
 *       * Neither the name of the <ORGANIZATION> nor the names of its         *
 *         contributors may be used to endorse or promote products derived     *
 *         from this software without specific prior written permission.       *
 *                                                                             *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" *
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE   *
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE  *
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE   *
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR         *
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF        *
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS    *
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN     *
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)     *
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE  *
 * POSSIBILITY OF SUCH DAMAGE.                                                 *
 ******************************************************************************/

/** Sample implementation of the marker list **/

#ifndef _JH_MARKER_LIST_H
#define _JH_MARKER_LIST_H

#include <stdlib.h>
#include "jh_datatypes.h"

#define JPEG_APP0   0xE0    /* APP0 marker code per JPEG spec*/

typedef struct jh_marker_node_ts {
    jh_marker_segment_t seg;
    struct jh_marker_node_ts* prev;
    struct jh_marker_node_ts* next;
} jh_marker_node_t;

typedef struct jh_marker_list_impl_ts {
    jh_marker_list_t pub;
    jh_marker_node_t* cur;
} jh_marker_list_impl_t;

/** interface implementation */
static int jh_insert_node(jh_marker_list_t* markers, unsigned char marker_id,
        unsigned int length1, const unsigned char *data1,
        unsigned int length2, const unsigned char *data2);
static int jh_get_prev(jh_marker_list_t* m_in, jh_marker_segment_t** segment);
static int jh_get_next(jh_marker_list_t* m_in, jh_marker_segment_t** segment);
static int jh_get(jh_marker_list_t* m_in, jh_marker_segment_t** segment);
static int jh_delete_segment(struct jh_marker_list_ts* markers, jh_marker_segment_t** segment);
static void jh_empty(struct jh_marker_list_ts* markers);

static void jh_init_marker_list(jh_marker_list_impl_t* markers) {
    jh_marker_list_t *iface = (jh_marker_list_t*)markers;
    iface->add = jh_insert_node;
    iface->jh_get_prev = jh_get_prev;
    iface->jh_get_next = jh_get_next;
    iface->get = jh_get;
    iface->jh_delete_segment = jh_delete_segment;
    iface->empty = jh_empty;
}

/* allocate a marker with the provided contents */
static jh_marker_node_t* jh_create_marker_node(unsigned char marker_id,
                                   unsigned length,
                                   unsigned char *data)
{
    jh_marker_node_t
        zero_marker = {{0}},
        *new_marker = (jh_marker_node_t *) jh_malloc(sizeof(jh_marker_node_t));
    if (NULL == new_marker) {
        return NULL;
    }
    *new_marker = zero_marker;
    new_marker->seg.data = (unsigned char *) jh_malloc(length * sizeof(unsigned char));
    if (NULL == new_marker->seg.data) {
        jh_free(new_marker);
        return NULL;
    }
    new_marker->seg.marker = marker_id;
    new_marker->seg.length = length;
    if (NULL != data) {
        memcpy(new_marker->seg.data, data, length * sizeof(unsigned char));
    }
    new_marker->next = NULL;
    new_marker->prev = NULL;
    return new_marker;
}

/* insert marker list */
static int jh_insert_node(jh_marker_list_t* base, unsigned char marker_id,
        unsigned int length1, const unsigned char *data1,
        unsigned int length2, const unsigned char *data2)
{
    jh_marker_list_impl_t* markers = (jh_marker_list_impl_t*) base;
    jh_marker_node_t* new_node = NULL;
    if((NULL == markers) || (NULL == data1) ||
            marker_id < JPEG_APP0
            || marker_id > JPEG_APP0 + 15) {
        return 0;
    }
    new_node = jh_create_marker_node(marker_id, length1 + length2, NULL);
    if (NULL == new_node) {
        return 0;
    }
    if (NULL != data1) {
        memcpy(new_node->seg.data, data1, length1);
        if (NULL != data2) {
            memcpy(new_node->seg.data + length1, data2, length2);
        }
    }
    if (NULL == markers->cur) { /* empty list */
        markers->cur = new_node;
        return 1;
    }
    if (NULL == markers->cur->next) { /* end of the list */
        markers->cur->next = new_node;
        new_node->prev = markers->cur;
        markers->cur = new_node;
        return 1;
    }
    new_node->next = markers->cur->next;
    new_node->prev = markers->cur;
    markers->cur->next->prev = new_node;
    markers->cur->next = new_node;
    markers->cur = new_node;
    return 1;
}


static int jh_get_prev(jh_marker_list_t* m_in, jh_marker_segment_t** segment) {
    jh_marker_list_impl_t* markers = (jh_marker_list_impl_t*)m_in;
    if (NULL == markers || NULL == markers->cur || NULL == markers->cur->prev ) {
        *segment = NULL;
        return 0;
    }
    *segment = (jh_marker_segment_t*)markers->cur->prev; /* we pass the node here */
    markers->cur = markers->cur->prev;
    return 1;
}

static int jh_get_next(jh_marker_list_t* m_in, jh_marker_segment_t** segment) {
    jh_marker_list_impl_t* markers = (jh_marker_list_impl_t*)m_in;
    if (NULL == markers || NULL == markers->cur || NULL == markers->cur->next ) {
        *segment = NULL;
        return 0;
    }
    *segment = (jh_marker_segment_t*)markers->cur->next;
    markers->cur = markers->cur->next;
    return 1;
}

static int jh_get(jh_marker_list_t* m_in, jh_marker_segment_t** segment) {
    jh_marker_list_impl_t* markers = (jh_marker_list_impl_t*)m_in;
    if (NULL == markers || NULL == markers->cur ) {
        *segment = NULL;
        return 0;
    }
    *segment = (jh_marker_segment_t*)markers->cur;
    return 1;
}


/**< @brief Deletes specified segment from the list
 *
 * @param[in|out] markers list of markers
 * @param[in|out] segment segment of interest
 * @return 1 if successful 0 if failed
 */
static int jh_delete_segment(struct jh_marker_list_ts* markers, jh_marker_segment_t** segment) {
    jh_marker_node_t *node = NULL == segment ? NULL : (jh_marker_node_t *)*segment;
    if (NULL == segment || NULL == *segment || NULL == markers) {
        return 0;
    }
    if (NULL == node->next && NULL == node->prev && ((jh_marker_list_impl_t*)markers)->cur == node ) { /* last node in the list*/
        ((jh_marker_list_impl_t*)markers)->cur = NULL; /* invalidate the current node */
    } else {
        if (NULL != node->next) { /* not end of the list */
            node->next->prev = node->prev;
            if (node == ((jh_marker_list_impl_t*)markers)->cur) {
                ((jh_marker_list_impl_t*)markers)->cur = node->next; /* set current to the next it will be corrected
                                                                if possible in following if statement */
            }
        }
        if (NULL != node->prev) { /* not beginning of the list */
            node->prev->next = node->next;
            if (node == ((jh_marker_list_impl_t*)markers)->cur) {
                ((jh_marker_list_impl_t*)markers)->cur = node->prev; /* after delete current node becomes previous unless beginning of the list */
            }
        }
    }
    jh_free(node->seg.data);
    jh_free(node);
    *segment = NULL;
    return 1;
}

/**< @brief Emtpy list of markers
 *
 * @param[in|out] markers list of markers
 */
static void jh_empty(struct jh_marker_list_ts* markers) {
    jh_marker_segment_t *seg = NULL;
    while(markers->jh_get_prev(markers, &seg));
    /* delete markers */
    markers->get(markers, &seg);
    do {
        markers->jh_delete_segment(markers, &seg); /* it is safe to delete from empty list */
    } while(markers->get(markers, &seg));
}

#endif /*_JH_MARKER_LIST_H*/
