/******************************************************************************
 * The copyright in this software is being made available under the BSD        *
 * License, included below. This software may be subject to other third party  *
 * and contributor rights, including patent rights, and no such rights are     *
 * granted under this license.                                                 *
 *    <OWNER> = ISO                                                            *
 *    <ORGANIZATION> = ISO                                                     *
 *    <YEAR> = 2015                                                            *
 *    Copyright (c) 2012-2015, ISO                                             *
 *                                                                             *
 * All rights reserved.                                                        *
 * Redistribution and use in source and binary forms, with or without          *
 * modification, are permitted provided that the following conditions are met: *
 *       * Redistributions of source code must retain the above copyright      *
 *         notice,this list of conditions and the following disclaimer.        *
 *       * Redistributions in binary form must reproduce the above             *
 *         copyright notice, this list of conditions and the following         *
 *         disclaimer in the documentation and/or other materials provided     *
 *         with the distribution.                                              *
 *       * Neither the name of the <ORGANIZATION> nor the names of its         *
 *         contributors may be used to endorse or promote products derived     *
 *         from this software without specific prior written permission.       *
 *                                                                             *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" *
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE   *
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE  *
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE   *
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR         *
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF        *
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS    *
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN     *
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)     *
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE  *
 * POSSIBILITY OF SUCH DAMAGE.                                                 *
 ******************************************************************************/

#include <setjmp.h>
#include <stdlib.h>
#include "jinclude.h"
#include "jpeglib.h"
#include "jerror.h"
#include "jh_helpers.h"
#include "jh_jpeg_wrap.h"
#include "jh_malloc.h"
#include "jh_stream_datadst.h"
#include "jh_stream_datasrc.h"


/* JPEG-HDR application marker # */
#define JH_APPM    (JH_JPEG_APP0 + 11)
/*- 2 byte length(data itself + 2byte length)*/

typedef struct {
    struct jpeg_error_mgr pub;	  /* "public" fields */
    char* buffer;                 /* at least jpg_get_max_message_length() bytes */
    jmp_buf setjmp_buffer;        /* for return to caller */
} error_mgr;

static void jh_jpg_error_exit(j_common_ptr ds) {
    error_mgr* err = (error_mgr*)ds->err;
    (*ds->err->format_message)(ds, err->buffer);
    longjmp(err->setjmp_buffer, 1);
}

static void jh_jpg_output_message(j_common_ptr ds) {
    error_mgr* err = (error_mgr*) ds->err;
    (*ds->err->format_message) (ds, err->buffer);
}

static void jh_jpg_init_err(j_common_ptr ds, error_mgr* err, char* message) {
    ds->err = jpeg_std_error(&err->pub);
    err->buffer = message;
    err->pub.error_exit = jh_jpg_error_exit;
    err->pub.output_message = jh_jpg_output_message;
}

/* Compression part */

static int jh_set_subsampling( struct jpeg_compress_struct* cinfo, int subsampling, jh_error_t* err) {
    jpeg_component_info * compptr = NULL;
    int ret = EXIT_SUCCESS;
#define JH_SET_COMP(index,id,hsamp,vsamp,quant,dctbl,actbl)  \
    (compptr = &cinfo->comp_info[index], \
    compptr->component_id = (id),        \
    compptr->h_samp_factor = (hsamp),    \
    compptr->v_samp_factor = (vsamp),    \
    compptr->quant_tbl_no = (quant),     \
    compptr->dc_tbl_no = (dctbl),        \
    compptr->ac_tbl_no = (actbl) )
    switch (subsampling) {
    case 444:
        JH_SET_COMP(0, 1, 1, 1, 0, 0, 0);
        JH_SET_COMP(1, 2, 1, 1, 1, 1, 1);
        JH_SET_COMP(2, 3, 1, 1, 1, 1, 1);
        break;
    case 420:
    case 0:
        JH_SET_COMP(0, 1, 2,2, 0, 0,0);
        JH_SET_COMP(1, 2, 1,1, 1, 1,1);
        JH_SET_COMP(2, 3, 1,1, 1, 1,1);
        break;
    case 422:
        JH_SET_COMP(0, 1, 2,1, 0, 0,0);
        JH_SET_COMP(1, 2, 1,1, 1, 1,1);
        JH_SET_COMP(2, 3, 1,1, 1, 1,1);
        break;
    default:
        jh_set_error(err, JH_INVALID_ARGUMENT, "Unsupported subsampling (%d) requested", subsampling);
        ret = EXIT_FAILURE;
    }
#undef JH_SET_COMP
    return ret;
}

static int jh_init_jpeg_engine(struct jpeg_compress_struct *cinfo, error_mgr *jerr, int qual,
                            int subsampling, const jh_image_t *image, jh_stream_t* stream, jh_error_t *err) {
    char message[JMSG_LENGTH_MAX] = {0};
    int ret = 1;
    jh_jpg_init_err((j_common_ptr)cinfo, jerr, message);
    if (setjmp(jerr->setjmp_buffer)) {
        error_mgr* jpg_err_mgr = (error_mgr*) cinfo->err;
        char msg[JH_MSG_SIZE] = "JPEG encoder error: ";
        strncat(msg, jpg_err_mgr->buffer, JH_MSG_SIZE - strlen(msg) );
        jh_set_error(err, JH_JPEG_ERROR, msg);
        return 0;
    }
    jpeg_create_compress(cinfo);
    jh_stream_dest(cinfo, stream);
    cinfo->image_width = (unsigned)image->columns;
    cinfo->image_height = (unsigned)image->rows;
    switch (image->pixel_type) {
    case JH_YUV444I8:
        cinfo->in_color_space = JCS_YCbCr;
        cinfo->input_components = 3;
        cinfo->jpeg_color_space = JCS_YCbCr;
        break;
    case JH_RGB8:
        cinfo->in_color_space = JCS_RGB;
        cinfo->input_components = 3;
        cinfo->jpeg_color_space = JCS_YCbCr;
        break;
    case JH_GRAY8:
        cinfo->in_color_space = JCS_GRAYSCALE;
        cinfo->input_components = 1;
        cinfo->jpeg_color_space = JCS_GRAYSCALE;
        break;
    default:
        jh_set_error(err, JH_INVALID_ARGUMENT, "jh_jpeg_compress: unknown input color space: %d",
                image->pixel_type);
        ret = 0;
        goto cleanup;
    }
    jpeg_set_defaults(cinfo);
    cinfo->optimize_coding = 1;
    if (jh_set_subsampling(cinfo, subsampling, err) == EXIT_FAILURE) {
        ret = 0;
        goto cleanup;
    }
    jpeg_set_quality(cinfo, qual, TRUE);
    jpeg_start_compress(cinfo, TRUE);
cleanup:
    return ret;
}

/* Define to dump the 8-bit SDR and subband images going into and coming out of the baseline JPEG library; useful for debugging. */
#undef DUMP_COMPONENT_IMAGES


#ifdef DUMP_COMPONENT_IMAGES
static void write_8bit_ppm(const char *fname, unsigned char *data, unsigned int w, unsigned int h, unsigned int channels ) {
    FILE *fp = fopen(fname, "wb");
    if(NULL == fp) {
        return;
    }
    fprintf(fp,"P%c\n%d %d\n255\n", ((3 == channels) ? '6' : '5'), w, h);
    fwrite((unsigned char*)data, sizeof(unsigned char), w * h * channels, fp);
    fclose(fp);
}
#endif

int jh_jpeg_compress_and_attach(const jh_image_t *sdri, int qual, int subsampling,
                                jh_marker_list_t *markers, jh_stream_t *strm,
                                jh_monitor_t * monitor, jh_error_t* err)
{
    struct jpeg_compress_struct	cinfo = {0};
    error_mgr jerr = {{0}};
    int	ret = 1;
    unsigned stride = 0;
    JSAMPLE * scanline = 0;
    jh_marker_segment_t *seg = NULL;
    
#ifdef DUMP_COMPONENT_IMAGES
    if(3 == jh_channel_count(sdri->pixel_type)) {
        write_8bit_ppm(sdri->pixel_type == JH_RGB8 ? "sdr_to_encode.ppm" : "residual_to_encode.ppm",
            (unsigned char *) sdri->data, sdri->columns, sdri->rows, 3);
    }
#endif
    /*sanity check*/
    if (NULL == sdri || NULL == sdri->data || 0 >= sdri->columns || 0 >= sdri->rows ) {
        jh_set_error(err, JH_INVALID_ARGUMENT, "jh_jpeg_compress_and_attach: Invalid image size or null pointer");
        return 0;
    }
    ret = jh_init_jpeg_engine(&cinfo, &jerr, qual, subsampling, sdri, strm, err);
    if (!ret) {
        goto cleanup;
    }
    /* rewind marker list */
    if( NULL != markers) {
        while (markers->jh_get_prev(markers, &seg));
        /* store markers */
        markers->get(markers, &seg);
        while (NULL != seg) {
            jpeg_write_marker(&cinfo, seg->marker, seg->data, seg->length);
            markers->jh_get_next(markers, &seg);
            if (jh_test_and_reset(&monitor->cancel)) {
                ret = 0;
                jh_set_error(err, JH_CANCELED, "Canceled by user");
                goto cleanup;
            }
        }
    }
    /*add base image*/
    scanline = (JSAMPLE *)sdri->data;
    stride = cinfo.image_width * cinfo.input_components;
    while(cinfo.next_scanline < cinfo.image_height) {
        jpeg_write_scanlines(&cinfo, &scanline, 1);
        scanline += stride;
        if (jh_test_and_reset(&monitor->cancel)) {
            ret = 0;
            jh_set_error(err, JH_CANCELED, "Canceled by user");
            goto cleanup;
        }
    }
    jpeg_finish_compress(&cinfo);
cleanup:
    jpeg_destroy_compress(&cinfo);
    return ret;
}


/*decompression part*/

typedef struct { /* JPEG decompression struct */
    struct  jpeg_decompress_struct cinfo;
    int     output_scanline;     /* current output scanline */
    jh_marker_list_t *marker_list;/* user defined container to store APP and COM markers */
    jh_error_t*   result;        /* the way to set a return code while parsing m11 segments */
    jh_monitor_t* monitor;       /* to watch monitor while parsing m11 segments */
} jpeghdr_decompress_struct;
typedef jpeghdr_decompress_struct* jh_decompress_ptr;


/* Functions need to support reading markers from the source */

/**<
 * @brief Read a byte from JPEG data source
 * @param[in|out] cinfo
 * @return byte read
 */
static unsigned char jh_read_next_byte (j_decompress_ptr cinfo) {
    if (cinfo->src->bytes_in_buffer == 0 &&
            !(*cinfo->src->fill_input_buffer)(cinfo))
        return EOF;
    cinfo->src->bytes_in_buffer--;
    return *cinfo->src->next_input_byte++;
}

/**<
 * @brief Read specified number of bytes from JPEG data source into a buffer
 * @param[in|out] cinfo
 * @param[in|out] buf  data buffer
 * @param[in] siz  buffer size
 * @return number of bytes read
 */
static size_t jh_read_bytes (j_decompress_ptr cinfo, unsigned char *buf, size_t siz) {
    size_t nbread = 0;
    while (cinfo->src->bytes_in_buffer < siz) {
        const size_t n = cinfo->src->bytes_in_buffer;
        if (n) {
            memcpy((void *)buf,
                (const void *)cinfo->src->next_input_byte, n);
            buf += n;
            siz -= n;
            nbread += n;
            cinfo->src->next_input_byte += n;
            cinfo->src->bytes_in_buffer = 0;
        }
        if (!(*cinfo->src->fill_input_buffer)(cinfo))
            return nbread;
    }
    memcpy((void *)buf, (const void *)cinfo->src->next_input_byte, siz);
    nbread += siz;
    cinfo->src->next_input_byte += siz;
    cinfo->src->bytes_in_buffer -= siz;
    return nbread;
}

/**<
 * @brief JPEG lib application markers and COM processing callback
 * @param[in|out] cinfo JPEG compression object
 * @return  1 if successful
 */
static boolean jh_marker_store(j_decompress_ptr cinfo) {
    jh_decompress_ptr jhinf = (jh_decompress_ptr)cinfo;
    jh_marker_list_t* markers = jhinf->marker_list;
    static unsigned char data[0xffff-2]; /* max data size in marker */
    unsigned size = jh_read_next_byte(cinfo) << 8;
    size |= jh_read_next_byte(cinfo);
    size -= 2;
    if ( 0 == size ) { /* empty marker */
        return TRUE;
    }
    if (jh_read_bytes(cinfo, data, size) != (size_t) size) { /* we copy data for simplicity sake. it can be done without copying */
        jh_set_error(jhinf->result, JH_JPEG_FORMAT_ERROR, "Wrong size of a marker segment.");
        return FALSE;
    }
    return markers->add(markers, cinfo->unread_marker, size, data, 0, NULL);
}

/* END of Functions need to support reading markers from the source */


int jh_jpeg_extract_markers(jh_stream_t*  strm, jh_marker_list_t* markers, jh_error_t *err) {
    jpeghdr_decompress_struct ds = {{0}};
    error_mgr jerr = {{0}};
    int ret = 1;
    int i = 0;
    jh_jpg_init_err((j_common_ptr)&ds, &jerr, err->description);
    if (setjmp(jerr.setjmp_buffer)) {
        err->code = JH_JPEG_ERROR;
        ret = 0;
        goto cleanup;
    }
    ds.marker_list = markers;
    jpeg_create_decompress(&ds.cinfo);
    jh_stream_src(&ds.cinfo, strm);
    for( i = 2; i < 16; i++) { /* don't store 0 and 1 markers */
        jpeg_set_marker_processor(&ds.cinfo, JPEG_APP0 + i, jh_marker_store);  /*callback for parsing app11*/
    }
    if (jpeg_read_header((struct jpeg_decompress_struct*)&ds, TRUE) != JPEG_HEADER_OK) {
        err->code = JH_JPEG_ERROR;
        goto cleanup;
    }
    for( i = 2; i < 16; i++) { /* don't store 0 and 1 markers */
        jpeg_save_markers(&ds.cinfo, JPEG_APP0 + i, 0);
    }
cleanup:
    jpeg_destroy_decompress(&ds.cinfo);
    return ret;
}

int jh_jpeg_extract_image(jh_stream_t*  strm, int scale_num, int scale_denom, jh_image_t* sdri, jh_error_t* err) {
    struct jpeg_decompress_struct ds = {0};
    error_mgr jerr = {{0}};
    jh_image_t zero_im = {0};
    unsigned char *pt = NULL;
    JSAMPLE *scanline = NULL;
    int ret = 1;
    int channels = 3;
    jh_jpg_init_err((j_common_ptr)&ds, &jerr, err->description);
    if (setjmp(jerr.setjmp_buffer)) {
        err->code = JH_JPEG_ERROR;
        goto cleanup;
    }
    jpeg_create_decompress(&ds);
    jh_stream_src(&ds, strm);
    if (jpeg_read_header(&ds, TRUE) != JPEG_HEADER_OK) {
        err->code = JH_JPEG_ERROR;
        goto cleanup;
    }
    switch (sdri->pixel_type) {
    case JH_YUV444I8:
        ds.out_color_space = JCS_YCbCr;
        break;
    case JH_RGB8:
        ds.out_color_space = JCS_RGB;
        break;
    case JH_GRAY8:
        ds.out_color_space = JCS_GRAYSCALE;
        channels = 1;
        break;
    default:
        jh_set_error(err, JH_INVALID_ARGUMENT, "JPEG wrapper was requested unknown pixel type %d", sdri->pixel_type);
        goto cleanup;
    }
    ds.scale_num = scale_num;
    ds.scale_denom = scale_denom;
    jpeg_start_decompress(&ds);
    sdri->data = (unsigned char*) jh_malloc(sizeof(unsigned char) * ds.output_width * ds.output_height * channels);
    pt = (unsigned char*) sdri->data;
    if (NULL == sdri->data) {
        jh_set_error(err, JH_MEMORY, "Could not allocate %d bytes in JPEG wrapper ", ds.output_width * ds.output_height * channels);
        goto cleanup;
    }
    sdri->columns = ds.output_width;
    sdri->rows = ds.output_height;
    sdri->stride = sdri->columns * channels;
    while (ds.output_scanline < ds.output_height) {
        scanline = pt + ds.output_scanline * ds.output_width * channels;
        if (!jpeg_read_scanlines(&ds, &scanline, 1)) {
            err->code = JH_JPEG_ERROR;
            goto cleanup;
        }
    }
cleanup:
    if (JH_OK != err->code) {
        jh_pixel_type_t pix = sdri->pixel_type;
        jh_free(pt);
        *sdri = zero_im;
        sdri->pixel_type = pix;
        ret = 0;
    }
    jpeg_destroy_decompress(&ds);
#ifdef DUMP_COMPONENT_IMAGES
    if(3 == jh_channel_count(sdri->pixel_type)) {
        write_8bit_ppm(sdri->pixel_type == JH_RGB8 ? "sdr_from_decode.ppm" : "residual_from_decode.ppm",
                       (unsigned char *) sdri->data, sdri->columns, sdri->rows, 3);
    }
#endif
    return ret;
}

int jh_jpeg_read_header(jh_stream_t* strm, int* width, int* height, int* channels, jh_error_t* ext_err) {
    struct jpeg_decompress_struct ds = {0};
    error_mgr err = {{0}};
    jh_jpg_init_err((j_common_ptr)&ds, &err, ext_err->description);
    if (setjmp(err.setjmp_buffer)) {
      jpeg_destroy_decompress(&ds);
      ext_err->code = JH_JPEG_ERROR;
      return 0;
    }
    jpeg_create_decompress(&ds);
    jh_stream_src(&ds, strm);
    jpeg_read_header(&ds, TRUE);
    *width = ds.image_width;
    *height = ds.image_height;
    *channels = ds.num_components;
    jpeg_destroy_decompress(&ds);
    return 1;
}

void jh_jpeg_release_image_data(unsigned char* data) {               /**< [in]    image data allocated by jh_jpeg_exctract_image */
        jh_free((void*)data);
}

