/******************************************************************************
 * The copyright in this software is being made available under the BSD        *
 * License, included below. This software may be subject to other third party  *
 * and contributor rights, including patent rights, and no such rights are     *
 * granted under this license.                                                 *
 *    <OWNER> = ISO                                                            *
 *    <ORGANIZATION> = ISO                                                     *
 *    <YEAR> = 2015                                                            *
 *    Copyright (c) 2012-2015, ISO                                             *
 *                                                                             *
 * All rights reserved.                                                        *
 * Redistribution and use in source and binary forms, with or without          *
 * modification, are permitted provided that the following conditions are met: *
 *       * Redistributions of source code must retain the above copyright      *
 *         notice,this list of conditions and the following disclaimer.        *
 *       * Redistributions in binary form must reproduce the above             *
 *         copyright notice, this list of conditions and the following         *
 *         disclaimer in the documentation and/or other materials provided     *
 *         with the distribution.                                              *
 *       * Neither the name of the <ORGANIZATION> nor the names of its         *
 *         contributors may be used to endorse or promote products derived     *
 *         from this software without specific prior written permission.       *
 *                                                                             *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" *
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE   *
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE  *
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE   *
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR         *
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF        *
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS    *
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN     *
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)     *
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE  *
 * POSSIBILITY OF SUCH DAMAGE.                                                 *
 ******************************************************************************/

/* implementation of destination manager in terms of jpeg-hdr stream */
/* Comments and function description are from IJG jpeg library sample */

/* this is not a core library module, so it doesn't define JPEG_INTERNALS */
#include "jinclude.h"
#include "jerror.h"
#include "jh_stream_datasrc.h"

/* Expanded data source object for stdio input */

typedef struct {
    struct jpeg_source_mgr pub;   /* public fields */
    jh_stream_t *strm;
} jh_stream_source_mgr;


/* prototypes */
static void jh_init_strm_source (j_decompress_ptr cinfo);
static boolean jh_fill_strm_input_buffer (j_decompress_ptr cinfo);
static void jh_skip_strm_input_data (j_decompress_ptr cinfo, long num_bytes);
static void jh_term_strm_source (j_decompress_ptr cinfo);


/*
 * Initialize source --- called by jpeg_read_header
 * before any data is actually read.
 */
static void jh_init_strm_source (j_decompress_ptr cinfo)
{
    jh_stream_source_mgr* src = (jh_stream_source_mgr*) cinfo->src;
    jh_stream_t *strm = src->strm;
    if (!strm->pri->on_init(strm, (unsigned char **) &src->pub.next_input_byte, &src->pub.bytes_in_buffer)) {
        ERREXIT1(cinfo, JERR_OUT_OF_MEMORY, 10);
    }
    return;
}


/*
 * Fill the input buffer --- called whenever buffer is emptied.
 *
 * In typical applications, this should read fresh data into the buffer
 * (ignoring the current state of next_input_byte & bytes_in_buffer),
 * reset the pointer & count to the start of the buffer, and return TRUE
 * indicating that the buffer has been reloaded.  It is not necessary to
 * fill the buffer entirely, only to obtain at least one more byte.
 *
 * There is no such thing as an EOF return.  If the end of the file has been
 * reached, the routine has a choice of ERREXIT() or inserting fake data into
 * the buffer.  In most cases, generating a warning message and inserting a
 * fake EOI marker is the best course of action --- this will allow the
 * decompressor to output however much of the image is there.  However,
 * the resulting error message is misleading if the real problem is an empty
 * input file, so we handle that case specially.
 *
 * In applications that need to be able to suspend compression due to input
 * not being available yet, a FALSE return indicates that no more data can be
 * obtained right now, but more may be forthcoming later.  In this situation,
 * the decompressor will return to its caller (with an indication of the
 * number of scanlines it has read, if any).  The application should resume
 * decompression after it has loaded more data into the input buffer.  Note
 * that there are substantial restrictions on the use of suspension --- see
 * the documentation.
 *
 * When suspending, the decompressor will back up to a convenient restart point
 * (typically the start of the current MCU). next_input_byte & bytes_in_buffer
 * indicate where the restart point will be if the current call returns FALSE.
 * Data beyond this point must be rescanned after resumption, so move it to
 * the front of the buffer rather than discarding it.
 */

static boolean jh_fill_strm_input_buffer (j_decompress_ptr cinfo)
{
    jh_stream_source_mgr* src = (jh_stream_source_mgr*) cinfo->src;
    jh_stream_t *strm = src->strm;
    if(!strm->pri->on_empty(strm, (unsigned char **) &src->pub.next_input_byte, &src->pub.bytes_in_buffer))
        ERREXIT1(cinfo, JERR_OUT_OF_MEMORY, 10);
    return TRUE;
}


/*
 * Skip data --- used to skip over a potentially large amount of
 * uninteresting data (such as an APPn marker).
 *
 * Writers of suspendable-input applications must note that skip_input_data
 * is not granted the right to give a suspension return.  If the skip extends
 * beyond the data currently in the buffer, the buffer can be marked empty so
 * that the next read will cause a fill_input_buffer call that can suspend.
 * Arranging for additional bytes to be discarded before reloading the input
 * buffer is the application writer's problem.
 */

static void jh_skip_strm_input_data (j_decompress_ptr cinfo, long num_bytes)
{
    jh_stream_source_mgr* src = (jh_stream_source_mgr*) cinfo->src;
    jh_stream_t *strm = src->strm;
    if(!strm->seek(strm, num_bytes, JH_SEEK_CUR))
        ERREXIT(cinfo, JERR_BUFFER_SIZE);
    src->pub.next_input_byte += num_bytes; /* HACK: it should be implemented in terms of stream */
    src->pub.bytes_in_buffer -= num_bytes; /* rather then directly manipulating the source mgr pointers */
    /* both skip and seek should be synchronized. Namely we supposed to use stream->offset
     * and stream->current-buf_size here. However it would require cast to the actual stream type thus
     * destroying polymorphism. Probably adding additional method returning needed data may solve the problem
     */
}


/*
 * An additional method that can be provided by data source modules is the
 * resync_to_restart method for error recovery in the presence of RST markers.
 * For the moment, this source module just uses the default resync method
 * provided by the JPEG library.  That method assumes that no backtracking
 * is possible.
 */


/*
 * Terminate source --- called by jpeg_finish_decompress
 * after all data has been read.  Often a no-op.
 *
 * NB: *not* called by jpeg_abort or jpeg_destroy; surrounding
 * application must deal with any cleanup that should happen even
 * for error exit.
 */

static void jh_term_strm_source (j_decompress_ptr cinfo)
{
    jh_stream_source_mgr* src = (jh_stream_source_mgr*) cinfo->src;
    jh_stream_t *strm = src->strm;
    if(!strm->pri->on_term(strm, src->pub.bytes_in_buffer))
        ERREXIT(cinfo, JERR_INPUT_EMPTY);
}



/*
 * Prepare for input from a supplied memory buffer.
 * The buffer must contain the whole JPEG data.
 */

void jh_stream_src (j_decompress_ptr cinfo, jh_stream_t *strm)
{
    jh_stream_source_mgr
            *src,
            src_zero = {{0}};
    if (strm == NULL) /* sanity check */
        ERREXIT(cinfo, JERR_BUFFER_SIZE); /* NOTE: we use this error code to indicate NULL stream */

    /* The source object is made permanent so that a series of JPEG images
     * can be read from the same buffer by calling jpeg_mem_src only before
     * the first one.
     */
    if (cinfo->src == NULL) { /* first time for this JPEG object? */
        cinfo->src = (struct jpeg_source_mgr *)
                        (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_PERMANENT,
                                SIZEOF(jh_stream_source_mgr));
    }
    src = (jh_stream_source_mgr*)cinfo->src;
    *src = src_zero;
    src->pub.init_source = jh_init_strm_source;
    src->pub.fill_input_buffer = jh_fill_strm_input_buffer;
    src->pub.skip_input_data = jh_skip_strm_input_data;
    src->pub.resync_to_restart = jpeg_resync_to_restart; /* use default method */
    src->pub.term_source = jh_term_strm_source;
    src->strm = strm;
}

