/******************************************************************************
 * The copyright in this software is being made available under the BSD        *
 * License, included below. This software may be subject to other third party  *
 * and contributor rights, including patent rights, and no such rights are     *
 * granted under this license.                                                 *
 *    <OWNER> = ISO                                                            *
 *    <ORGANIZATION> = ISO                                                     *
 *    <YEAR> = 2015                                                            *
 *    Copyright (c) 2012-2015, ISO                                             *
 *                                                                             *
 * All rights reserved.                                                        *
 * Redistribution and use in source and binary forms, with or without          *
 * modification, are permitted provided that the following conditions are met: *
 *       * Redistributions of source code must retain the above copyright      *
 *         notice,this list of conditions and the following disclaimer.        *
 *       * Redistributions in binary form must reproduce the above             *
 *         copyright notice, this list of conditions and the following         *
 *         disclaimer in the documentation and/or other materials provided     *
 *         with the distribution.                                              *
 *       * Neither the name of the <ORGANIZATION> nor the names of its         *
 *         contributors may be used to endorse or promote products derived     *
 *         from this software without specific prior written permission.       *
 *                                                                             *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" *
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE   *
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE  *
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE   *
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR         *
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF        *
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS    *
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN     *
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)     *
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE  *
 * POSSIBILITY OF SUCH DAMAGE.                                                 *
 ******************************************************************************/

#include <iostream>
#include <fstream>
#include <cmath>
#include <sys/stat.h>
#include "json.h"
#include "parameters_parser.h"
#include "tonemap.h"
#include "jh_helpers.h"



/*
 Parameter files use the JSON format.  Parameters can appear
 anywhere in the file.  Order doesn't matter unless duplicates are present, in which case
 an extracted parameter will take on the last value encountered.  Lines are key-value pairs,
 where the keys have been selected to match their corresponding C struct element names.  
 All parameters are optional; default values will be supplied by the parser for any 
 mandatory parameters that are not present in the JSON file.
 
 Note that JSON does not support comments; if you want to include them, just use key-value
 pairs like this:
	"_my_comment_key_" : "Y U No Allow Comments"
 
 See http://json.org/ for documentation of the JSON format.
*/

#define VALIDATE_ERROR(err, return_value)  if(null == err) { return (return_value); }
#define CHECK_ERROR(err) if(JH_OK != err->code) { return 1; }

/* return 0 on success, 1 on error. */
static int ParseJSON(const char *paramsfile, Json::Value& root, jh_error_t *err)
{
    int ret = 0;
    try {
        VALIDATE_ERROR(err, 1);
        Json::Reader reader;
        std::ifstream strm(paramsfile, std::ifstream::binary);
        if (!strm.good()) {
            jh_set_error(err, JH_ERROR, "Cannot open parameters file %s.", paramsfile);
            ret = 1;
        } else if( !reader.parse(strm, root, false) ) {
            /* report the failure location in the document. */
            jh_set_error(err, JH_ERROR, "Error parsing JSON parameters file: %s", reader.getFormattedErrorMessages().c_str());
            ret = 1;
        }
    }
    catch (...) {
        jh_set_error(err, JH_ERROR, "Error parsing JSON parameters file.", paramsfile);
        ret = 1;
    }
    return ret;
}

static int ReadJSONFloatArray(const Json::Value& jsonArrayObject, float *array, const unsigned int numElements, jh_error_t *err)
{
    try {
        VALIDATE_ERROR(err, 1);
        if(jsonArrayObject.isNull()) {
            /* object not present */
            return 0;
        }
        if(numElements != jsonArrayObject.size()) {
            /* error: NULL jsonValue (parameter not present, incorrect key used, etc.) or array size from 
             *        file doesn't match expected size. Caller should provide default array in this case.
             */
            jh_set_error(err, JH_ERROR, "Improperly sized float array encountered.");
            return 1;
        }
        for(unsigned int index=0; index<jsonArrayObject.size(); ++index) {
            array[index] = jsonArrayObject[index].asFloat();
        }
    }
    catch (...) {
        jh_set_error(err, JH_ERROR, "Error reading JSON float array.");
        return 1;
    }
    return 0;
}

static int ReadJSONIntWithDefault(const Json::Value& jsonValue, const int defaultValue, jh_error_t *err)
{
    try {
        VALIDATE_ERROR(err, defaultValue);
        if(jsonValue.isNull()) {
            return defaultValue;
        }
        int intValue = jsonValue.asInt();
        return intValue;
    }
    catch (...) {
        jh_set_error(err, JH_ERROR, "Error reading JSON int value.");
    }
    return defaultValue;
}

static bool ReadJSONBoolWithDefault(const Json::Value& jsonValue, const bool defaultValue, jh_error_t *err)
{
    try {
        VALIDATE_ERROR(err, defaultValue);
        if(jsonValue.isNull()) {
            return defaultValue;
        }
        bool boolValue = jsonValue.asBool();
        return boolValue;
    }
    catch (...) {
        jh_set_error(err, JH_ERROR, "Error reading JSON Boolean value.");
    }
    return defaultValue;
}

static float ReadJSONFloatWithDefault(const Json::Value& jsonValue, const float defaultValue, jh_error_t *err)
{
    try {
        VALIDATE_ERROR(err, defaultValue);
        if(jsonValue.isNull()) {
            return defaultValue;
        }
        float floatValue = jsonValue.asFloat();
        return floatValue;
    }
    catch (...) {
        jh_set_error(err, JH_ERROR, "Error reading JSON float value.");
    }
    return defaultValue;
}

int GetEncoderParameters(jh_encoder_params_t* encoderParams, const char *paramsfile, jh_error_t *err)
{
    try {
        VALIDATE_ERROR(err, 1);
        Json::Value root;
        if( ParseJSON(paramsfile, root, err) ) {
            return 1;
        }

        /* fill the encoder parameters with the JSON contents */
        encoderParams->base_quality = ReadJSONIntWithDefault(root["base_quality"], encoderParams->base_quality, err);
        CHECK_ERROR(err);
        encoderParams->subband_quality = ReadJSONIntWithDefault(root["subband_quality"], encoderParams->subband_quality, err);
        CHECK_ERROR(err);
        encoderParams->samp2nits = ReadJSONFloatWithDefault(root["samp2nits"], encoderParams->samp2nits, err);
        CHECK_ERROR(err);
        encoderParams->noise_floor = ReadJSONFloatWithDefault(root["noise_floor"], encoderParams->noise_floor, err);
        CHECK_ERROR(err);
        encoderParams->predictive_coding = ReadJSONIntWithDefault(root["predictive_coding"], encoderParams->predictive_coding, err);
        CHECK_ERROR(err);
        encoderParams->base_subsampling = ReadJSONIntWithDefault(root["base_subsampling"], encoderParams->base_subsampling, err);
        CHECK_ERROR(err);
        encoderParams->subband_subsampling = ReadJSONIntWithDefault(root["subband_subsampling"], encoderParams->subband_subsampling, err);
        CHECK_ERROR(err);
        
        /* array default values preserved if no tag exists */
        ReadJSONFloatArray(root["chroma_minmax"], encoderParams->chroma_minmax, 4, err);
        CHECK_ERROR(err);
        ReadJSONFloatArray(root["swb"], encoderParams->swb, 3, err);
        CHECK_ERROR(err);
        ReadJSONFloatArray(root["spr"], encoderParams->spr, 9, err);
        CHECK_ERROR(err);
        ReadJSONFloatArray(root["cs_conv_mat"], encoderParams->cs_conv_mat, JH_CS_MATRIX_SIZE, err);
        CHECK_ERROR(err);
        ReadJSONFloatArray(root["base_cs_trc"], encoderParams->base_cs_trc, JH_CS_TRC_SIZE, err);
        CHECK_ERROR(err);
        ReadJSONFloatArray(root["cs_resi_mat"], encoderParams->cs_resi_mat, JH_CS_MATRIX_SIZE, err);
        CHECK_ERROR(err);
        ReadJSONFloatArray(root["target_cs_mat"], encoderParams->target_cs_mat, JH_CS_MATRIX_SIZE, err);
        CHECK_ERROR(err);
    }
    catch (...) {
        jh_set_error(err, JH_ERROR, "Error reading encoder parameters.");
        return 1;
    }
    return 0;
}

int GetDecoderParameters(jh_decoder_params_t* decoderParams, const char *paramsfile, jh_error_t *err)
{
    try {
        VALIDATE_ERROR(err, 1);
        Json::Value root;
        if( ParseJSON(paramsfile, root, err) ) {
            return 1;
        }

        /* fill the decoder parameters with the JSON contents */
        decoderParams->num = ReadJSONIntWithDefault(root["num"], decoderParams->num, err);
        CHECK_ERROR(err);
        decoderParams->denom = ReadJSONIntWithDefault(root["denom"], decoderParams->denom, err);
        CHECK_ERROR(err);
    }
    catch (...) {
        jh_set_error(err, JH_INVALID_ARGUMENT, "Error reading decoder parameters from file %s", paramsfile);
        return 1;
    }
    return 0;
}

int GetTonemappingParameters(tmo_params_t* tmoParams, const char *paramsfile, jh_error_t *err)
{
    try {
        VALIDATE_ERROR(err, 1);
        Json::Value root;
        if( ParseJSON(paramsfile, root, err) ) {
            return 1;
        }
        
        /* fill the decoder parameters with the JSON contents */
        tmoParams->ec = ReadJSONFloatWithDefault(root["exposure_compensation"], 0.0f, err);
        CHECK_ERROR(err);
    }
    catch (...) {
        jh_set_error(err, JH_ERROR, "Error reading tonemapping parameters.");
        return 1;
    }
    return 0;
}

/* Fill decoder parameters with default values, for cases when no parameters file is given. */
void GetDefaultDecoderParameters(jh_decoder_params_t* decoderParams, jh_error_t *err)
{
    if(null == err) {
        return;
    }
    jh_set_error(err, JH_OK, "No error.");
    jh_decoder_set_default(decoderParams, err);
    return;
}

/* Fill tonemapping operator parameters with default values, for cases when no parameters file is given. */
void GetDefaultTonemappingParameters(tmo_params_t* tmoParams, jh_error_t *err)
{
    if(null == err) {
        return;
    }
    jh_set_error(err, JH_OK, "No error.");
    tonemap_set_default(tmoParams, err);
    return;
}

const std::string GetStringParameterWithDefault(const char *paramsfile, const char *key, const char *defaultValue, jh_error_t *err)
{
    try {
        VALIDATE_ERROR(err, defaultValue);
        Json::Value root;
        if( ParseJSON(paramsfile, root, err) ) {
            return defaultValue;
        }
        if(0 == root[key].asString().length()) {
            return defaultValue;
        }
        return(root[key].asString());
    }
    catch (...) {
        jh_set_error(err, JH_ERROR, "Error reading JSON string value.");
    }
    return defaultValue;
}

int GetICCProfile(unsigned char **icc_profile_data, size_t *icc_profile_size, const char *paramsfile, jh_error_t *err)
{
    VALIDATE_ERROR(err, 1);
    const std::string profile_path = GetStringParameterWithDefault(paramsfile, "base_icc_profile_path", "", err);
    CHECK_ERROR(err);
    const char *icc_profile_path = profile_path.c_str();
    if(profile_path.empty()) {
        /* No profile specified in the parameters file; that's OK */
        return 0;
    }
    *icc_profile_data = NULL;
    *icc_profile_size = 0;
    
    /* get the profile file size */
    struct stat stat_buf;
    off_t profileSize = (stat(icc_profile_path, &stat_buf) == 0) ? stat_buf.st_size : -1;
    if(profileSize <= 0) {
        jh_set_error(err, JH_ERROR, "Zero-length or non-existent profile %s specified in parameters file.\n", icc_profile_path);
        return 1;
    }
    
    FILE *fp = fopen(icc_profile_path, "rb");
    if(NULL == fp) {
        jh_set_error(err, JH_ERROR, "Failed to open profile %s; not added to JPEG stream.\n", icc_profile_path);
        return 1;
    }
    
    *icc_profile_size = static_cast<size_t>(profileSize);
    *icc_profile_data = (unsigned char *) malloc(*icc_profile_size * sizeof(unsigned char));
    if(NULL == *icc_profile_data) {
        jh_set_error(err, JH_ERROR, "Failed to allocate memory for profile %s; not added to JPEG stream.\n", icc_profile_path);
        fclose(fp);
        return 1;
    }
 
    if(*icc_profile_size != fread(*icc_profile_data, sizeof(unsigned char), *icc_profile_size, fp)) {
        jh_set_error(err, JH_ERROR, "Error reading profile data for %s; not added to JPEG stream.\n", icc_profile_path);
        fclose(fp);
        return 1;
    }
    fclose(fp);
    
    return 0;
}

#undef VALIDATE_ERROR
#undef CHECK_ERROR
