/******************************************************************************
 * The copyright in this software is being made available under the BSD        *
 * License, included below. This software may be subject to other third party  *
 * and contributor rights, including patent rights, and no such rights are     *
 * granted under this license.                                                 *
 *    <OWNER> = ISO                                                            *
 *    <ORGANIZATION> = ISO                                                     *
 *    <YEAR> = 2015                                                            *
 *    Copyright (c) 2012-2015, ISO                                             *
 *                                                                             *
 * All rights reserved.                                                        *
 * Redistribution and use in source and binary forms, with or without          *
 * modification, are permitted provided that the following conditions are met: *
 *       * Redistributions of source code must retain the above copyright      *
 *         notice,this list of conditions and the following disclaimer.        *
 *       * Redistributions in binary form must reproduce the above             *
 *         copyright notice, this list of conditions and the following         *
 *         disclaimer in the documentation and/or other materials provided     *
 *         with the distribution.                                              *
 *       * Neither the name of the <ORGANIZATION> nor the names of its         *
 *         contributors may be used to endorse or promote products derived     *
 *         from this software without specific prior written permission.       *
 *                                                                             *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" *
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE   *
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE  *
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE   *
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR         *
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF        *
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS    *
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN     *
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)     *
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE  *
 * POSSIBILITY OF SUCH DAMAGE.                                                 *
 ******************************************************************************/

#include <string.h>
#include <stdio.h>
#include <math.h>
#include "jh_helpers.h"
#include "jh_datatypes.h"
#include "jh_color.h"
#include "jh_malloc.h"
#include "jh_log.h"
#include "tonemap.h"

typedef struct tmo_uhist_t {  /*! uniformly spaced histogram */
    int     nbins;
    int     npixels;
    float   pMinMax[2];
    float*  pData;
    float   gamma;
} tmo_uhist_t;

static char tmo_tonemapper_canceled[] = "Tonemapper canceled";
static tmo_uhist_t* tmo_create_histogram(const size_t cnbins, float fMin, float fMax);
static void tmo_delete_histogram(tmo_uhist_t* im);
static jh_result_t tmo_calc_histogram(const jh_image_t* hdri, tmo_uhist_t *hist);
static inline void* tmo_get_address_lum(void* base, int columns, int column, int row) {
    return (void*)((char*)base + (row * columns + column) * sizeof(float));
}

static int tmo_exgamma(const tmo_params_t* p, jh_monitor_t* monitor, const jh_image_t* hdri,
                          jh_image_t* sdri, jh_error_t* error);

static int is_invalid_tmo_params(const tmo_params_t * p, jh_error_t *err, int columns, int rows)
{
    if (p == NULL) {
        jh_set_error(err, JH_INVALID_ARGUMENT, "NULL tmo parameters");
        return 1;
    }
    /* check parameter ranges */
    if ((p->ec < -4.0f) || (p->ec > 4.0f)) {
        jh_set_error(err, JH_INVALID_ARGUMENT, "ec value %f is out of range [-4, 4]", p->ec);
        return 1;
    }
    return 0;
}


int tonemap(const tmo_params_t* p, jh_monitor_t* monitor, const jh_image_t* hdri, jh_image_t* sdri, jh_error_t* error)
{
    int ret = 1;
    if (jh_is_invalid_error(error) || is_invalid_tmo_params(p, error, hdri->columns, hdri->rows)
            || jh_is_invalid_image(hdri, error) || jh_is_invalid_image(sdri, error)
            || jh_is_null(monitor, "monitor", error)) {
        return 0;
    }
    
    if(JH_RGBF == hdri->pixel_type) {
        if(0 == tmo_exgamma(p, monitor, hdri, sdri, error)) {
            return 0;
        }
    }
    else if(JH_GRAYF == hdri->pixel_type) {
        int i = 0, j = 0;
        const unsigned int RGB = 3;
        jh_image_t *hdriRGB = NULL, *sdriRGB = NULL;
        
        /* tmo_expgamma doesn't support gray images; workaround here by converting gray <-> RGB for tonemapping. */
        hdriRGB = jh_create_image(hdri->columns, hdri->rows, JH_RGBF);
        if(NULL == hdriRGB) {
            ret = 0;
            goto cleanup;
        }
        for(i=0; i<hdriRGB->rows; i++) {
            float *grayPtr = (float *) jh_get_row(hdri, i);
            float *rgbPtr = (float *) jh_get_row(hdriRGB, i);
            for(j=0; j<hdriRGB->columns; j++) {
                *rgbPtr++ = *grayPtr;
                *rgbPtr++ = *grayPtr;
                *rgbPtr++ = *grayPtr++;
            }
        }
        sdriRGB = jh_create_image(sdri->columns, sdri->rows, JH_RGB8);
        if(NULL == sdriRGB) {
            ret = 0;
            goto cleanup;
        }
        if(0 == tmo_exgamma(p, monitor, hdriRGB, sdriRGB, error)) {
            ret = 0;
            goto cleanup;
        }
   
        /* convert the RGB SDR image back gray */
        for(i=0; i<sdriRGB->rows; i++) {
            unsigned char *grayPtr = (unsigned char *) jh_get_row(sdri, i);
            unsigned char *rgbPtr = (unsigned char *) jh_get_row(sdriRGB, i);
            for(j=0; j<sdriRGB->columns; j++) {
                float tmp = jh_hdr_lum(rgbPtr[0], rgbPtr[1], rgbPtr[2], NULL) + 0.5f;   /* 0.5f in anticipation of cast to uchar */
                *grayPtr++ = (tmp <= 0.0f) ? 0 : ((tmp >= 255.0f) ? 255 : (unsigned char) tmp);
                rgbPtr += RGB;
            }
        }
    cleanup:
        jh_delete_image(hdriRGB);
        jh_delete_image(sdriRGB);
    }

    return ret;
}

int tonemap_set_default(tmo_params_t *tp, jh_error_t *err)
{
    if (NULL == err) {
        return 0;
    }
    if (NULL == tp) {
        jh_set_error(err, JH_INVALID_ARGUMENT, "tonemap_set_default: null pointer passed as parameter");
        return 0;
    }
    tp->ec = 0.0f;
    return 1;
}

static float tmo_calc_exposure_compensation(const jh_image_t* src, int nch);
static int tmo_exgamma_main_loop(jh_monitor_t* monitor, const jh_image_t* hdri,
                                    jh_image_t* sdri, jh_error_t* error,
                                    float ec, unsigned char LUT[], int channels);

static int tmo_exgamma(const tmo_params_t* p, jh_monitor_t* monitor, const jh_image_t* hdri,
                          jh_image_t* sdri, jh_error_t* error) {
    int i, channels,
        ret = 1;
    float ec, scale;
    unsigned char LUT[1<<JH_BITS];
    channels = 3; /*Only JH_RGBF supported*/
    if (JH_RGBF != hdri->pixel_type) {
        jh_set_error(error, JH_INVALID_ARGUMENT, "Exposure-gamma tonemapper only supports floating-point HDR images");
        ret = 0;
        goto cleanup;
    }
    ec = tmo_calc_exposure_compensation(hdri, channels);
    ec += p->ec; /* add manually entered exposure compensation */
    for (i = 0; i<(1<<JH_BITS); ++i) { /*Populate LUT*/
        LUT[i] = (unsigned char)(255.f*jh_gamma_encode((float)i/((1<<JH_BITS)-1)) + .5);
    }
    scale = powf(2.f, ec) * ((1<<JH_BITS)-1) * .8f; /* 0.8 is a bit less then half stop dynamic range extension (0.5 is too much)*/
    ret = tmo_exgamma_main_loop(monitor, hdri, sdri, error, scale, LUT, channels);
cleanup:
    return ret;
}

static int tmo_exgamma_main_loop(jh_monitor_t* monitor, const jh_image_t* hdri,
                                    jh_image_t* sdri, jh_error_t* error,
                                    float scale, unsigned char LUT[], int channels) {
    unsigned char
        *d0 = (unsigned char*)sdri->data,
        *d  = (unsigned char*)sdri->data;
    unsigned int
        ds = sdri->stride;
    int i, row = 0,
        lastrow = sdri->rows,
        have_loglum = (NULL != sdri->loglum);
    float
        *ratio = (float *) sdri->loglum,
        lymin = 10000000.f,
        lymax = -10000000.f,
        gamma_lut[256];
    if (have_loglum) {
        /* calc gamma corrected tmi to linear float*/
        for (i = 0; i < 256; i++ ) {
            gamma_lut[i] = jh_gamma_decode(i/255.f) + 1.f/(1<<JH_BITS);
        }
    }

    while(row < lastrow) {
        int columns = sdri->columns;
        float *s = (float*)jh_get_row(hdri, row);
        while(columns--){
            d[0] = LUT[jh_round(s[0] * scale)];
            d[1] = LUT[jh_round(s[1] * scale)];
            d[2] = LUT[jh_round(s[2] * scale)];
            if(have_loglum) {
                float lum = jh_hdr_lum(s[0], s[1], s[2], NULL);
                float lum_tm = jh_hdr_lum(gamma_lut[d[0]], gamma_lut[d[1]], gamma_lut[d[2]], NULL);
                if((lum < JH_MIN_LUM) || (lum_tm < JH_MIN_LUM)) {
                    *ratio = 1.0f;
                } else {
                    *ratio = lum/lum_tm;
                }
                if( *ratio < lymin ) { lymin = *ratio; }
                if( *ratio > lymax ) { lymax = *ratio; }
                ++ratio;
            }
            d += 3;
            s += 3;
        }
        d = d0 = d0 + ds;
        if (monitor && jh_test_and_reset(&monitor->cancel)) {
            jh_set_error(error, JH_CANCELED, tmo_tonemapper_canceled);
            return 0;
        }
        row++;
    }
    if (have_loglum) {
        *ratio++ = lymin;    /* store the min & max at the end of loglum */
        *ratio   = lymax;
    }
    return 1;
}

static float tmo_calc_exposure_compensation(const jh_image_t* src, int nch)
{
    float    ec = 0.f,           /*ec based on rectrangle*/
             min_ec = 0.f,       /*limiting does not allow ec to adjust white into grey*/
             max_ec = 0.f;       /*limiting does not allow ec to adjust black into grey*/
    float    inverse_step = 0.f, hist_min = 0.f, lum_min = 0.f, lum_max = 0.f;
    tmo_uhist_t* loc_hist;
    tmo_uhist_t* g_hist;
    int      row, lastrow, ss, skip;
    if (NULL == src || 3 != nch) {
        return 0.f;
    }
    loc_hist = tmo_create_histogram(JH_HIST_SIZE, JH_LN_JH_MIN_LUM, JH_LN_MAX_LUM); /*local, rectangle based histogram*/
    if (NULL == loc_hist) {
        return 0.f;
    }
    g_hist = tmo_create_histogram(JH_HIST_SIZE, JH_LN_JH_MIN_LUM, JH_LN_MAX_LUM);   /*global, whole image  histogram*/
    if (NULL == g_hist) {
        tmo_delete_histogram(loc_hist);
        return 0.f;
    }
    inverse_step = ((float)(loc_hist->nbins-1))/(loc_hist->pMinMax[1] - loc_hist->pMinMax[0]);
    hist_min = loc_hist->pMinMax[0];
    lum_min = expf(hist_min);
    lum_max = expf(loc_hist->pMinMax[1]);
    /*Calculate local histogram*/
    row = 0;
    lastrow = src->rows;
    if (src->loglum) { /*We have a table of logarithm luminance*/
        float loglum_min = jh_log(lum_min);
        float loglum_max = jh_log(lum_max);
        while(row < lastrow) {
            int column = src->columns;
            float *lumptr = (float*)tmo_get_address_lum(src->loglum, src->columns, 0, row);
            while(column--){
                float loglum = *lumptr++;
                if(loglum <= loglum_min) {
                    loc_hist->pData[0]++;
                } else if(loglum >= loglum_max) {
                    loc_hist->pData[loc_hist->nbins-1]++;
                }
                else {
                    loc_hist->pData[(unsigned int)((loglum - hist_min) * inverse_step)]++;
                }
            }
            row++;
        }
    } else {
        while(row < lastrow) {
            int columns = src->columns;
            float *s = (float*)jh_get_row(src, row);
            while(columns--){
                float lum = jh_hdr_lum(*s, *(s+1), *(s+2), NULL);
                if(lum <= lum_min) {
                    loc_hist->pData[0]++;
                } else if(lum >= lum_max) {
                    loc_hist->pData[loc_hist->nbins-1]++;
                }
                else {
                    loc_hist->pData[(unsigned int)((jh_log(lum) - hist_min) * inverse_step)]++;
                }
                s += 3;
            }
            row++;
        }
    }
    jh_cumulative_sum(loc_hist->pData, JH_HIST_SIZE);
    loc_hist->npixels = src->columns * src->rows;
    /*Find median bin for local histogram*/
    ss = 0;
    while((loc_hist->pData[ss++] < ((loc_hist->npixels)>>1)) && (ss < loc_hist->nbins))
        ;
    /*Find the luminance corresponding the median bin and recalculate ec to match mid-gray*/
    ec = jh_log2(0.18f/expf((float)ss/inverse_step + hist_min));
    /*Calculate global histogram*/
    skip = 4;
    row = 0;
    lastrow = src->rows;
    if (src->loglum) { /*We have a table of logarithm luminance*/
        float loglum_min = jh_log(lum_min);
        float loglum_max = jh_log(lum_max);
        while(row < lastrow) {
            int column = 0;
            int last_column = src->columns;
            float *lumptr = (float*)tmo_get_address_lum(src->loglum, src->columns, column, row);
            while(column < last_column) {
                float loglum = *lumptr;
                ++g_hist->npixels;
                if(loglum <= loglum_min) {
                    g_hist->pData[0]++;
                } else if(loglum >= loglum_max) {
                    g_hist->pData[g_hist->nbins-1]++;
                } else {
                    g_hist->pData[(unsigned int)((loglum - hist_min) * inverse_step)]++;
                }
                column += skip;
                lumptr += skip;
            }
            row += skip;
        }
    } else {
        while(row < lastrow) {
            int column = 0;
            int last_column = src->columns;
            float *s = (float*)jh_get_row(src, row);
            while(column < last_column) {
                float lum = jh_hdr_lum(*s, *(s+1), *(s+2), NULL);
                ++g_hist->npixels;
                if(lum <= lum_min) {
                    g_hist->pData[0]++;
                } else if(lum >= lum_max) {
                    g_hist->pData[g_hist->nbins-1]++;
                } else {
                    g_hist->pData[(unsigned int)((jh_log(lum) - hist_min) * inverse_step)]++;
                }
                column += skip;
                s += 3 * skip;
            }
            row += skip;
        }
    }
    jh_cumulative_sum(g_hist->pData, JH_HIST_SIZE);
    ss = g_hist->nbins;
    /* Find bin corresponding to white (99% of the histogram) */
    while((g_hist->pData[--ss]) > (int)(.99f * (float)g_hist->npixels)) {
       /* empty */ ;
    }
    /* Find exposure compensation to bring 98% white to display white. 0.8 is heuristic */
    min_ec = jh_log2(.7f/expf((float)ss/inverse_step + hist_min));
    /* Find bin corresponding to black (2% of the histogram) */
    while((g_hist->pData[--ss]) > (int)(.02 * (float)g_hist->npixels)) {
        /* empty */ ;
    }
    max_ec = jh_log2(.02f/expf((float)ss/inverse_step + hist_min));
    tmo_delete_histogram(loc_hist);
    tmo_delete_histogram(g_hist);
    if (ec < min_ec) {
        ec = min_ec;
    } else if (ec > max_ec) {
        ec = max_ec;
    }
    return ec;
}

static tmo_uhist_t* tmo_create_histogram(const size_t cnbins, float fMin, float fMax)
{
    tmo_uhist_t* h;
    if (0 == cnbins || fMax <= fMin || fMax - fMin > JH_MAX_DR  /* empty image */
        || NULL == (h = (tmo_uhist_t*)malloc(sizeof(tmo_uhist_t)))) /* no memory */
    {
        return NULL;
    }
    if (NULL != (h->pData = (float*)jh_zalloc(cnbins * sizeof(float)))) {
        h->nbins = (int)cnbins;
        h->pMinMax[0] = fMin; h->pMinMax[1] = fMax;
        h->npixels = 0;
    }
    else {
        free((void*)h);
        h = NULL;
    }
    return h;
}

static void tmo_delete_histogram(tmo_uhist_t* im) {
    if(NULL == im) {
        return;
    }
    jh_free_aligned((void*) im->pData);
    im->pData = NULL;
    jh_free_aligned(im);
}

static jh_result_t tmo_calc_histogram(const jh_image_t* hdri, tmo_uhist_t *hist) {
    int i = 0;
    const int
        cols = NULL == hdri ? 0 : hdri->columns,
        rows = NULL == hdri ? 0 : hdri->rows;
    size_t pixels = 0;
    const float
        inverse_step = NULL == hist ? 0 : ((float)(hist->nbins - 1)) / (hist->pMinMax[1] - hist->pMinMax[0]),
        min_lum = NULL == hist ? 0 : expf(hist->pMinMax[0]),
        max_lum = NULL == hist ? 0 : expf(hist->pMinMax[1]);
    float *pTmp = NULL;
    float cum_lum = 0.f;
    float log_lum;

    if (NULL == hdri || NULL == hist) {
        return JH_INVALID_ARGUMENT;
    }
    /* init historgramm */
    for (pTmp = hist->pData, i=0; i < hist->nbins; i++) {
        *pTmp++ = 0;
    }
    pixels = 0;
    switch (hdri->pixel_type) {
    case JH_RGBF: {
        float *pin = (float*)hdri->data;
        for (i = 0; i < cols * rows; i += 16) {
            float Y = jh_hdr_lum(*pin, *(pin+1), *(pin+2), NULL);
            ++pixels;
            if (Y < min_lum) {
                hist->pData[0]++;
                log_lum = hist->pMinMax[0];
            } else if (Y > max_lum){
                hist->pData[JH_HIST_SIZE-1]++;
                log_lum = hist->pMinMax[1];
            } else {
                int ind;
                log_lum = jh_log(Y);
                ind = (int)((log_lum - hist->pMinMax[0]) * inverse_step);
                hist->pData[ind]++;
            }
            cum_lum += log_lum;
            pin += 3 * 16;
        }
        break;
    }
    default:
        return JH_ERROR;
    }
    hist->npixels = (int)pixels;
    hist->gamma = 1.7117f * expf(cum_lum / (int)pixels) + 0.6388f;
    return JH_OK;
}
